from django.apps import apps
from django.core import exceptions
from django.utils.translation import gettext_lazy as _

from hashids import Hashids
from rest_framework import fields

from hashid_field.conf import settings
from hashid_field.hashid import Hashid
from hashid_field.lookups import _is_int_representation


class UnconfiguredHashidSerialField(fields.Field):
    def bind(self, field_name, parent):
        super().bind(field_name, parent)
        raise exceptions.ImproperlyConfigured(
            "The field '{field_name}' on {parent} must be explicitly declared when used with a ModelSerializer".format(
                field_name=field_name, parent=parent.__class__.__name__))


class HashidSerializerMixin(object):
    usage_text = "Must pass a HashidField, HashidAutoField or 'app_label.model.field'"
    default_error_messages = {
        'invalid': _("value must be a positive integer or a valid Hashids string."),
        'invalid_hashid': _("'{value}' value must be a valid Hashids string."),
    }

    def __init__(self, **kwargs):
        self.hashid_salt = kwargs.pop('salt', settings.HASHID_FIELD_SALT)
        self.hashid_min_length = kwargs.pop('min_length', settings.HASHID_FIELD_MIN_LENGTH)
        self.hashid_alphabet = kwargs.pop('alphabet', settings.HASHID_FIELD_ALPHABET)
        self.allow_int_lookup = kwargs.pop('allow_int_lookup', settings.HASHID_FIELD_ALLOW_INT_LOOKUP)
        self.prefix = kwargs.pop('prefix', "")
        self._hashids = kwargs.pop('hashids', None)

        source_field = kwargs.pop('source_field', None)
        if source_field:
            from hashid_field import HashidField, BigHashidField, HashidAutoField, BigHashidAutoField
            if isinstance(source_field, str):
                try:
                    app_label, model_name, field_name = source_field.split(".")
                except ValueError:
                    raise ValueError(self.usage_text)
                model = apps.get_model(app_label, model_name)
                source_field = model._meta.get_field(field_name)
            elif not isinstance(source_field, (HashidField, BigHashidField, HashidAutoField, BigHashidAutoField)):
                raise TypeError(self.usage_text)
            self.hashid_salt = source_field.salt
            self.hashid_min_length = source_field.min_length
            self.hashid_alphabet = source_field.alphabet
            self.allow_int_lookup = source_field.allow_int_lookup
            self.prefix = source_field.prefix
            self._hashids =source_field._hashids
        if not self._hashids:
            self._hashids = Hashids(salt=self.hashid_salt, min_length=self.hashid_min_length,
                                    alphabet=self.hashid_alphabet)
        super().__init__(**kwargs)

    def to_internal_value(self, data):
        value = super().to_internal_value(data)
        try:
            return Hashid(value, salt=self.hashid_salt, min_length=self.hashid_min_length,
                          alphabet=self.hashid_alphabet, prefix=self.prefix, hashids=self._hashids)
        except ValueError:
            self.fail('invalid_hashid', value=data)


class HashidSerializerCharField(HashidSerializerMixin, fields.CharField):
    def to_representation(self, value):
        return str(value)

    def to_internal_value(self, data):
        hashid = super().to_internal_value(data)
        if isinstance(data, int) and not self.allow_int_lookup:
            self.fail('invalid_hashid', value=data)
        if isinstance(data, str) and not self.allow_int_lookup:
            # Make sure int lookups are not allowed, even if prefixed, unless the
            # given value is actually a hashid made up entirely of numbers.
            without_prefix = data[len(self.prefix):]
            if _is_int_representation(without_prefix) and without_prefix != hashid.hashid:
                self.fail('invalid_hashid', value=data)
        return hashid


class HashidSerializerIntegerField(HashidSerializerMixin, fields.IntegerField):
    def to_representation(self, value):
        return int(value)




from django import forms
from django.core import exceptions, checks
from django.core import validators as django_validators
from django.db import models
from django.db.models import Field
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from django.contrib.admin import widgets as admin_widgets
from hashids import Hashids

from .lookups import HashidExactLookup, HashidIterableLookup
from .lookups import HashidGreaterThan, HashidGreaterThanOrEqual, HashidLessThan, HashidLessThanOrEqual
from .descriptor import HashidDescriptor
from .hashid import Hashid
from .conf import settings
from .validators import HashidMaxValueValidator, HashidMinValueValidator


def _alphabet_unique_len(alphabet):
    return len([x for i, x in enumerate(alphabet) if alphabet.index(x) == i])


class HashidFieldMixin(object):
    default_error_messages = {
        'invalid': _("'%(value)s' value must be a positive integer or a valid Hashids string."),
        'invalid_hashid': _("'%(value)s' value must be a valid Hashids string."),
    }
    exact_lookups = ('exact', 'iexact', 'contains', 'icontains')
    iterable_lookups = ('in',)
    passthrough_lookups = ('isnull',)
    comparison_lookups = {
        'gt': HashidGreaterThan,
        'gte': HashidGreaterThanOrEqual,
        'lt': HashidLessThan,
        'lte': HashidLessThanOrEqual,
    }

    def __init__(self, salt=settings.HASHID_FIELD_SALT,
                 min_length=settings.HASHID_FIELD_MIN_LENGTH,
                 alphabet=settings.HASHID_FIELD_ALPHABET,
                 allow_int_lookup=settings.HASHID_FIELD_ALLOW_INT_LOOKUP,
                 enable_hashid_object=settings.HASHID_FIELD_ENABLE_HASHID_OBJECT,
                 enable_descriptor=settings.HASHID_FIELD_ENABLE_DESCRIPTOR,
                 prefix="", *args, **kwargs):
        self.salt = salt
        self.min_length = min_length
        self.alphabet = alphabet
        if _alphabet_unique_len(self.alphabet) < 16:
            raise exceptions.ImproperlyConfigured("'alphabet' must contain a minimum of 16 unique characters")
        self._hashids = Hashids(salt=self.salt, min_length=self.min_length, alphabet=self.alphabet)
        self.allow_int_lookup = allow_int_lookup
        self.enable_hashid_object = enable_hashid_object
        self.enable_descriptor = enable_descriptor
        self.prefix = prefix
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs['min_length'] = self.min_length
        kwargs['alphabet'] = self.alphabet
        kwargs['prefix'] = self.prefix
        return name, path, args, kwargs

    def check(self, **kwargs):
        errors = super().check(**kwargs)
        errors.extend(self._check_alphabet_min_length())
        errors.extend(self._check_salt_is_set())
        return errors

    def _check_alphabet_min_length(self):
        if _alphabet_unique_len(self.alphabet) < 16:
            return [
                checks.Error(
                    "'alphabet' must contain a minimum of 16 unique characters",
                    hint="Add more unique characters to custom 'alphabet'",
                    obj=self,
                    id='HashidField.E001',
                )
            ]
        return []

    def _check_salt_is_set(self):
        if self.salt is None or self.salt == "":
            return [
                checks.Warning(
                    "'salt' is not set",
                    hint="Pass a salt value in your field or set settings.HASHID_FIELD_SALT",
                    obj=self,
                    id="HashidField.W001",
                )
            ]
        return []

    @cached_property
    def validators(self):
        if self.enable_hashid_object:
            return super().validators
        else:
            # IntegerField will add min and max validators depending on the database we're connecting to, so we need
            # to override them with our own validator that knows how to `clean` the value before doing the check.
            validators_ = super().validators
            validators = []
            for validator_ in validators_:
                if isinstance(validator_, django_validators.MaxValueValidator):
                    validators.append(HashidMaxValueValidator(self, validator_.limit_value, validator_.message))
                elif isinstance(validator_, django_validators.MinValueValidator):
                    validators.append(HashidMinValueValidator(self, validator_.limit_value, validator_.message))
                else:
                    validators.append(validator_)
            return validators

    def encode_id(self, id):
        hashid = self.get_hashid(id)
        if self.enable_hashid_object:
            return hashid
        else:
            return str(hashid)

    def get_hashid(self, id):
        return Hashid(id, salt=self.salt, min_length=self.min_length, alphabet=self.alphabet,
                      prefix=self.prefix, hashids=self._hashids)

    def from_db_value(self, value, expression, connection):
        if value is None:
            return value
        return self.encode_id(value)

    def get_lookup(self, lookup_name):
        if lookup_name in self.exact_lookups:
            return HashidExactLookup
        if lookup_name in self.iterable_lookups:
            return HashidIterableLookup
        if lookup_name in self.comparison_lookups:
            return self.comparison_lookups[lookup_name]
        if lookup_name in self.passthrough_lookups:
            return super().get_lookup(lookup_name)
        return None  # Otherwise, we don't allow lookups of this type

    def to_python(self, value):
        if isinstance(value, Hashid):
            return value
        if value is None:
            return value
        try:
            hashid = self.encode_id(value)
        except ValueError:
            raise exceptions.ValidationError(
                self.error_messages['invalid'],
                code='invalid',
                params={'value': value},
            )
        return hashid

    def get_prep_value(self, value):
        if value is None or value == '':
            return None
        if isinstance(value, Hashid):
            return value.id
        try:
            hashid = self.get_hashid(value)
        except ValueError:
            raise ValueError(self.error_messages['invalid'] % {'value': value})
        return hashid.id

    def contribute_to_class(self, cls, name, **kwargs):
        super().contribute_to_class(cls, name, **kwargs)
        # if callable(self.prefix):
        #     self.prefix = self.prefix(field_instance=self, model_class=cls, field_name=name, **kwargs)
        if self.enable_descriptor:
            descriptor = HashidDescriptor(field_name=self.attname, salt=self.salt, min_length=self.min_length,
                                          alphabet=self.alphabet, prefix=self.prefix, hashids=self._hashids,
                                          enable_hashid_object=self.enable_hashid_object)
            setattr(cls, self.attname, descriptor)


class HashidCharFieldMixin:
    def formfield(self, **kwargs):
        defaults = {'form_class': forms.CharField}
        defaults.update(kwargs)
        if defaults.get('widget') == admin_widgets.AdminIntegerFieldWidget:
            defaults['widget'] = admin_widgets.AdminTextInputWidget
        if defaults.get('widget') == admin_widgets.AdminBigIntegerFieldWidget:
            defaults['widget'] = admin_widgets.AdminTextInputWidget
        # noinspection PyCallByClass,PyTypeChecker
        return Field.formfield(self, **defaults)


class HashidField(HashidFieldMixin, HashidCharFieldMixin, models.IntegerField):
    description = "A Hashids obscured IntegerField"


class BigHashidField(HashidFieldMixin, HashidCharFieldMixin, models.BigIntegerField):
    description = "A Hashids obscured BigIntegerField"

    def __init__(self, min_length=settings.HASHID_FIELD_BIG_MIN_LENGTH, *args, **kwargs):
        super().__init__(min_length=min_length, *args, **kwargs)


class HashidAutoField(HashidFieldMixin, models.AutoField):
    description = "A Hashids obscured AutoField"


class BigHashidAutoField(HashidFieldMixin, models.AutoField):
    # This inherits from AutoField instead of BigAutoField so that DEFAULT_AUTO_FIELD doesn't throw an error
    description = "A Hashids obscured BigAutoField"

    def get_internal_type(self):
        return 'BigAutoField'

    def rel_db_type(self, connection):
        return models.BigIntegerField().db_type(connection=connection)

    def __init__(self, min_length=settings.HASHID_FIELD_BIG_MIN_LENGTH, *args, **kwargs):
        super().__init__(min_length=min_length, *args, **kwargs)


# Monkey patch Django REST Framework, if it's installed, to throw exceptions if fields aren't explicitly defined in
# ModelSerializers. Not doing so can lead to hard-to-debug behavior.
try:
    from rest_framework.serializers import ModelSerializer
    from hashid_field.rest import UnconfiguredHashidSerialField

    ModelSerializer.serializer_field_mapping[HashidField] = UnconfiguredHashidSerialField
    ModelSerializer.serializer_field_mapping[BigHashidField] = UnconfiguredHashidSerialField
    ModelSerializer.serializer_field_mapping[HashidAutoField] = UnconfiguredHashidSerialField
    ModelSerializer.serializer_field_mapping[BigHashidAutoField] = UnconfiguredHashidSerialField
except ImportError:
    pass



from django.core.validators import MaxValueValidator, MinValueValidator


class HashidMaxValueValidator(MaxValueValidator):
    def __init__(self, hashid_field, limit_value, message=None):
        self.hashid_field = hashid_field
        super().__init__(limit_value, message)

    def clean(self, x):
        return self.hashid_field.get_prep_value(x)


class HashidMinValueValidator(MinValueValidator):
    def __init__(self, hashid_field, limit_value, message=None):
        self.hashid_field = hashid_field
        super().__init__(limit_value, message)

    def clean(self, x):
        return self.hashid_field.get_prep_value(x)



from functools import total_ordering

from hashids import Hashids


def _is_uint(candidate):
    """Returns whether a value is an unsigned integer."""
    return isinstance(candidate, int) and candidate >= 0


def _is_str(candidate):
    """Returns whether a value is a string."""
    return isinstance(candidate, str)


@total_ordering
class Hashid(object):
    def __init__(self, value, salt="", min_length=0, alphabet=Hashids.ALPHABET, prefix="", hashids=None):
        self._salt = salt
        self._min_length = min_length
        self._alphabet = alphabet
        self._prefix = str(prefix)

        # If hashids is provided, it's for optimization only, and should be initialized with the same salt, min_length
        # and alphabet, or else we will run into problems
        self._hashids = hashids or Hashids(salt=self._salt, min_length=self._min_length, alphabet=self._alphabet)
        if not self._valid_hashids_object():
            raise Exception("Invalid hashids.Hashids object")

        if value is None:
            raise ValueError("id must be a positive integer or a valid Hashid string")

        # Check if `value` is an integer and encode it.
        # This presumes hashids will only ever be strings, even if they are made up entirely of numbers
        if _is_uint(value):
            self._id = value
            self._hashid = self.encode(value)
        elif _is_str(value):
            # Verify that it begins with the prefix, which could be the default ""
            if value.startswith(self._prefix):
                value = value[len(self._prefix):]
            else:
                # Check if the given string is all numbers, and encode without requiring the prefix.
                # This is to maintain backwards compatibility, specifically being able to enter numbers in an admin.
                # If a hashid is typed in that happens to be all numbers, without the prefix, then it will be
                # interpreted as an integer and encoded (again).
                try:
                    value = int(value)
                except (TypeError, ValueError):
                    raise ValueError("value must begin with prefix {}".format(self._prefix))

            # Check if this string is a valid hashid, even if it's made up entirely of numbers
            _id = self.decode(value)
            if _id is None:
                # The given value is not a hashids string, so see if it's a valid string representation of an integer
                try:
                    value = int(value)
                except (TypeError, ValueError):
                    raise ValueError("value must be a positive integer or a valid Hashid string")

                # Make sure it's positive
                if not _is_uint(value):
                    raise ValueError("value must be a positive integer")

                # We can use it as-is
                self._id = value
                self._hashid = self.encode(value)
            else:
                # This is a valid hashid
                self._id = _id
                self._hashid = value
        elif isinstance(value, int) and value < 0:
            raise ValueError("value must be a positive integer")
        else:
            raise ValueError("value must be a positive integer or a valid Hashid string")

    @property
    def id(self):
        return self._id

    @property
    def hashid(self):
        return self._hashid

    @property
    def prefix(self):
        return self._prefix

    @property
    def hashids(self):
        return self._hashids

    def encode(self, id):
        return self._hashids.encode(id)

    def decode(self, hashid):
        ret = self._hashids.decode(hashid)
        if len(ret) == 1:
            return ret[0]
        else:
            return None

    def _valid_hashids_object(self):
        # The hashids.Hashids class randomizes the alphabet and pulls out separators and guards, thus not being
        # reversible. So all we can test is that the length of the alphabet, separators and guards are equal to the
        # original alphabet we gave it. We could also check that all of the characters we gave it are present, but that
        # seems excessive... this test will catch most errors.
        return self._salt == self._hashids._salt \
            and self._min_length == self._hashids._min_length \
            and len(self._alphabet) == len(self._hashids._alphabet + self._hashids._separators + self._hashids._guards)

    def __repr__(self):
        return "Hashid({}): {}".format(self._id, str(self))

    def __str__(self):
        return self._prefix + self._hashid

    def __int__(self):
        return self._id

    def __long__(self):
        return int(self._id)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (
                self._id == other._id and
                self._hashid == other._hashid and
                self._prefix == other._prefix
            )
        if isinstance(other, str):
            return str(self) == other
        if isinstance(other, int):
            return int(self) == other
        return NotImplemented

    def __lt__(self, other):
        if isinstance(other, self.__class__):
            return self._id < other._id
        if isinstance(other, type(self._id)):
            return self._id < other
        return NotImplemented

    def __len__(self):
        return len(str(self))

    def __hash__(self):
        return hash(str(self))

    def __getstate__(self):
        return self._id, self._salt, self._min_length, self._alphabet, self._prefix, self._hashid

    def __setstate__(self, state):
        self._id, self._salt, self._min_length, self._alphabet, self._prefix, self._hashid = state

    def __add__(self, other):
        return self._id + other

    def __sub__(self, other):
        return self._id - other

    def __mul__(self, other):
        return self._id * other

    def __matmul__(self, other):
        raise NotImplementedError("Hashid does not support matrix multiplication")

    def __truediv__(self, other):
        return self._id / other

    def __floordiv__(self, other):
        return self._id // other

    def __mod__(self, other):
        return self._id % other

    def __divmod__(self, other):
        return self.__floordiv__(other), self.__mod__(other)

    def __pow__(self, other, *args):
        return pow(self._id, other, *args)

    def __lshift__(self, other):
        return self._id << other

    def __rshift__(self, other):
        return self._id >> other

    def __and__(self, other):
        return self._id & other

    def __xor__(self, other):
        return self._id ^ other

    def __or__(self, other):
        return self._id | other
